//{{NO_DEPENDENCIES}}
// Microsoft Developer Studio generated include file.
// Used by WinElvis.rc
//
#define IDR_MENU1                       101
#define IDM_ELVIS                       101
#define IDI_ELVIS                       103
#define IDC_RIGHT_ARROW                 104
#define IDC_RIGHTARROW                  104
#define IDA_ELVIS                       106
#define IDD_DIR_SELECT                  107
#define IDD_ABOUT                       108
#define IDD_BUFFERS                     109
#define IDT_ELVIS                       110
#define IDD_GOTO                        112
#define IDD_SEARCH                      113
#define IDD_RUN                         114
#define IDD_TAGS                        115
#define IDD_REPLACE                     116
#define IDD_OPT_BUFFER                  117
#define IDD_OPT_USER                    118
#define IDD_OPT_WINDOW                  119
#define IDD_OPT_SYNTAX                  120
#define IDD_OPT_GLOBAL                  121
#define IDD_OPT_GUI                     122
#define IDI_ICON2                       126
#define IDAB_PHYS_MEM                   1000
#define IDAB_DISK_SPACE                 1001
#define IDBL_BUFFERS                    1002
#define IDGT_LINE                       1003
#define IDSR_IGNORECASE                 1004
#define IDSR_MAGIC                      1005
#define IDSR_AUTOSELECT                 1006
#define IDSR_WRAPSCAN                   1007
#define IDSR_EDCOMPATIBLE               1008
#define IDSR_FORWARD                    1009
#define IDSR_STRING                     1011
#define IDRUN_CMDLINE                   1012
#define IDTAG_FILES                     1013
#define IDTAG_USEQUEST                  1014
#define IDTAG_LINENUMBERS               1015
#define IDTAG_STATIC                    1016
#define IDTAG_INLINE                    1017
#define IDTAG_TYPEDEFS                  1018
#define IDTAG_VARIABLE                  1019
#define IDTAG_REFS                      1020
#define IDTAG_APPEND                    1021
#define IDREP_STR1                      1022
#define IDREP_STR2                      1023
#define IDREP_FROM                      1024
#define IDREP_TO                        1025
#define IDREP_MAGIC                     1029
#define IDREP_IGNORECASE                1030
#define IDREP_G                         1031
#define IDREP_P                         1032
#define IDREP_X                         1033
#define IDC_BO_AUTOINDENT               1034
#define IDC_BO_AUTOTAB                  1035
#define IDC_BO_BINARY                   1036
#define IDC_BO_MODIFIED                 1037
#define IDC_BO_SHIFTWIDTH               1038
#define IDC_BO_TABSTOPSTR               1039
#define IDC_BO_TABSTOP                  1040
#define IDC_BO_LINELENGHT               1041
#define IDC_BO_TEXTWIDTH                1041
#define IDC_BO_UNDOLEVELS               1042
#define IDC_BO_PARAGRAPHS               1043
#define IDC_BO_SECTIONS                 1044
#define IDC_BO_INPUTTAB                 1045
#define IDC_BO_CCPRG                    1046
#define IDC_BO_EQUALPRG                 1047
#define IDC_KWD                         1048
#define IDC_BO_KEYWORDPRG               1049
#define IDC_BO_MAKEPRG                  1050
#define IDC_UO_A                        1051
#define IDC_UO_B                        1052
#define IDC_WO_LIST                     1052
#define IDC_UO_C                        1053
#define IDC_WO_SHOWCMD                  1053
#define IDC_UO_D                        1054
#define IDC_WO_SHOWMATCH                1054
#define IDC_UO_E                        1055
#define IDC_WO_SHOWSTACK                1055
#define IDC_UO_F                        1056
#define IDC_WO_NUMBER                   1056
#define IDC_UO_G                        1057
#define IDC_WO_WRAP                     1057
#define IDC_UO_H                        1058
#define IDC_WO_COLUMNS                  1058
#define IDC_UO_I                        1059
#define IDC_WO_LINES                    1059
#define IDC_UO_J                        1060
#define IDC_WO_SCROLL                   1060
#define IDC_UO_K                        1061
#define IDC_WO_SIDESCROLL               1061
#define IDC_UO_L                        1062
#define IDC_WO_WRAPMARGIN               1062
#define IDC_SO_ELEMENT                  1064
#define IDC_SO_NORMAL                   1065
#define IDC_SO_BOLD                     1066
#define IDC_SO_EMPHASIZED               1067
#define IDC_SO_ITALIC                   1068
#define IDC_SO_UNDERLINED               1069
#define IDC_SO_FIXED                    1070
#define IDC_GO_AUTOPRINT                1071
#define IDC_GO_AUTOSELECT               1072
#define IDC_GO_AUTOWRITE                1073
#define IDC_GO_ANYERROR                 1074
#define IDC_GO_BACKUP                   1075
#define IDC_UO_M                        1076
#define IDC_GO_BEAUTIFY                 1076
#define IDC_UO_N                        1077
#define IDC_GO_DEFAULTREADONLY          1077
#define IDC_UO_O                        1078
#define IDC_GO_DIGRAPH                  1078
#define IDC_UO_P                        1079
#define IDC_GO_EDCOMPATIBLE             1079
#define IDC_UO_Q                        1080
#define IDC_GO_ERRORBELLS               1080
#define IDC_UO_R                        1081
#define IDC_GO_EXRC                     1081
#define IDC_GO_FLASH                    1082
#define IDC_GO_GDEFAULT                 1083
#define IDC_GO_IGNORECASE               1084
#define IDC_UO_S                        1085
#define IDC_GO_MESG                     1085
#define IDC_UO_T                        1086
#define IDC_GO_MAGIC                    1086
#define IDC_UO_U                        1087
#define IDC_GO_MODELINE                 1087
#define IDC_UO_V                        1088
#define IDC_GO_NOVICE                   1088
#define IDC_UO_W                        1089
#define IDC_GO_OPTIMIZE                 1089
#define IDC_UO_X                        1090
#define IDC_GO_PROMPT                   1090
#define IDC_UO_Y                        1091
#define IDC_GO_REMAP                    1091
#define IDC_UO_Z                        1092
#define IDC_GO_SAFER                    1092
#define IDC_GO_SHOWMARKUPS              1093
#define IDC_GO_SYNC                     1094
#define IDC_GO_TAGSTACK                 1095
#define IDC_GO_TERSE                    1096
#define IDC_GO_VERBOSE                  1097
#define IDC_GO_WARN                     1098
#define IDC_GO_WARNINGBELLS             1099
#define IDC_GO_WRAPSCAN                 1100
#define IDC_GO_WRITEANY                 1101
#define IDC_GO_MODELINES                1102
#define IDC_GO_NEARSCROLL               1103
#define IDC_GO_REPORT                   1104
#define IDC_GO_TAGLENGTH                1107
#define IDC_GO_DIRECTORY                1109
#define IDC_GO_HOME                     1110
#define IDC_GO_NONASCII                 1111
#define IDC_GO_SHELL                    1112
#define IDC_GO_TAGS                     1113
#define IDC_GUI_SCROLLBAR               1114
#define IDC_GUI_TOOLBAR                 1115
#define IDC_GUI_FONTS                   1116
#define IDC_GUI_BOLD                    1117
#define IDC_GUI_ITALIC                  1118
#define IDC_GUI_UNDERLINED              1119
#define IDC_BO_READEOL                  1128
#define IDC_GO_WRITEEOL                 1130
#define IDAB_VERSION                    1131
#define IDC_GUI_BOXED                   1132
#define IDC_GUI_GRAPHIC                 1133
#define IDC_GUI_FIXED                   1134
#define IDC_GUI_FOREGROUND              1135
#define IDC_GUI_MENUBAR                 1136
#define IDC_GUI_STATUSBAR               1137
#define IDC_GUI_BACKGROUND              1138
#define IDC_PROPORTIONAL                1139
#define IDC_GUI_PROPORTIONAL            1139
#define IDM_FILE_EXIT                   40001
#define IDM_FILE_NEW                    40002
#define IDM_FILE_OPEN                   40003
#define IDM_FILE_SAVE                   40004
#define IDM_FILE_SAVEAS                 40005
#define IDM_FILE_CD                     40006
#define IDM_FILE_PRINT                  40007
#define IDM_FILE_PRINTSETUP             40008
#define IDM_HELP_INDEX                  40009
#define IDM_HELP_EX                     40010
#define IDM_HELP_VI                     40011
#define IDM_HELP_OPTIONS                40012
#define IDM_HELP_ABOUT                  40013
#define IDM_EDIT_UNDO                   40014
#define IDM_EDIT_REDO                   40015
#define IDM_EDIT_CUT                    40016
#define IDM_EDIT_COPY                   40017
#define IDM_EDIT_PASTE                  40018
#define IDM_WINDOW_NEW                  40019
#define IDM_WINDOW_NEXT                 40020
#define IDM_WINDOW_PREVIOUS             40021
#define IDM_WINDOW_NFILE                40022
#define IDM_WINDOW_PFILE                40023
#define IDM_WINDOW_BUFFER               40024
#define IDM_SEARCH_SEARCH               40026
#define IDM_SEARCH_AGAIN                40027
#define IDM_SEARCH_REPLACE              40028
#define IDM_SEARCH_GOTO                 40029
#define IDM_SEARCH_NEXT_ERROR           40030
#define IDM_TOOLS_COMPILE               40034
#define IDM_TOOLS_MAKE                  40035
#define IDM_TOOLS_TAGS                  40036
#define IDM_TOOLS_RUN                   40037
#define IDM_TOOLS_SHELL                 40038
#define IDM_OPTIONS_SAVE                40039
#define IDM_OPTIONS_FONT                40040
#define IDM_OPTIONS_TOOLBAR             40041
#define IDM_OPTIONS_SCROLLBAR           40042
#define IDM_FILE_SPLIT                  40043
#define IDM_OPTIONS_BUFFER              40044
#define IDM_OPTIONS_USER                40045
#define IDM_OPTIONS_WINDOW              40046
#define IDM_OPTIONS_SYNTAX              40047
#define IDM_OPTIONS_GLOBAL              40048
#define IDM_OPTIONS_GUI                 40049
#define IDC_STATIC                      -1

// Next default values for new objects
// 
#ifdef APSTUDIO_INVOKED
#ifndef APSTUDIO_READONLY_SYMBOLS
#define _APS_NEXT_RESOURCE_VALUE        127
#define _APS_NEXT_COMMAND_VALUE         40050
#define _APS_NEXT_CONTROL_VALUE         1140
#define _APS_NEXT_SYMED_VALUE           101
#endif
#endif
