//{{NO_DEPENDENCIES}}
// Microsoft Developer Studio generated include file.
// Used by wintags.rc
//
#define IDD_TAGS                        101
#define ID_CUR_FILE                     1000
#define ID_CUR_TAGS                     1001
#define ID_TOTAL_TAGS                   1002
#define IDC_STATIC                      -1

// Next default values for new objects
// 
#ifdef APSTUDIO_INVOKED
#ifndef APSTUDIO_READONLY_SYMBOLS
#define _APS_NEXT_RESOURCE_VALUE        102
#define _APS_NEXT_COMMAND_VALUE         40001
#define _APS_NEXT_CONTROL_VALUE         1003
#define _APS_NEXT_SYMED_VALUE           101
#endif
#endif
