
/* Copyright 2000 by Dr. David Alan Gilbert (elvis@treblig.org)      */
/*   based on guix11.c, with help from the gnome-hello code          */
/*   the Glade code, guys on the gnome-devel list, gmc,              */
/*   and the numbers 42 and 56.                                      */
/* It may be distributed and used under the same license terms       */
/* as the rest of the Elvis package which it is part of.             */

/* Initial crack at this - probably will move it over to being a descendent
   of desktop-editor or something similar */
module Elvis {
  /* An interface just to be used within Elvis */
  interface ElvisPrivate {
    /* Open a file in a new window, relative to a given directory */
    void openFile (in string dir, in string toopen);
  };
};
